#ifndef HELL_INTEGRATION_HH
#define HELL_INTEGRATION_HH

#include "types.hh"

namespace HELL {

  double integration_gauss_inf(double f(double,void*), double xmin, double prec, double *error, void *par, int *status=NULL);
  double integration_gauss(double f(double,void*), double xmin, double xmax, double prec, double *error, void *par, int *status=NULL);

  // SImpson method for integration
  class SimpsonIntegration {
  public:
    static Real integrate(Real fct_vals[], Real a, Real b, unsigned n,
                          Real * err = 0);
  private:
    static Real simpson_rule_(Real fct_vals[], Real a, Real b, unsigned n);
  };

  
  // Mellin stuff

  // e^(- I phi) /. phi -> 3/4 Pi
  const Complex MELLIN_EIPHI = Complex(-0.707106781186547524400844362105,
                                        0.707106781186547524400844362105);

  const Real MELLIN_Z_MAX_PARAM =  40.0;

  inline Real inv_mellin_max_z(Real x)
  {
    return MELLIN_Z_MAX_PARAM / log(1/x);
  }
  
  
} // namespace hell

#endif // HELL_INTEGRATION_HH
