#ifndef HELL_TYPES_HH
#define HELL_TYPES_HH

#include <iostream>
#include <cmath>
#include <complex>

namespace HELL {

  // types
  typedef double Real;
  typedef std::complex<Real> Complex;

  const Real EPSILON = 1e-11;

  // constants
  const Real PI = M_PI;

    // Imaginary unit
  const Complex I(Real(0),Real(1));


  // Functions

  // Put log in hell-namespace
  inline Real log(Real x)
  {
    return std::log(x);
  }
  inline Complex log(const Complex & z)
  {
    return std::log(z);
  }
  inline Real log10(Real x)
  {
    return std::log10(x);
  }
  inline Complex log10(const Complex & z)
  {
    return std::log10(z);
  }

  // Put exp in hell-namespace
  inline Real exp(Real x)
  {
    return std::exp(x);
  }
  inline Complex exp(const Complex & z)
  {
    return std::exp(z);
  }
  
  // Put real in hell-namespace
  inline Real real(const Complex & z)
  {
    return std::real(z);
  }
  inline Real real(Real x)
  {
    return x;
  }

  // Put imag in hell-namespace
  inline Real imag(const Complex & z)
  {
    return std::imag(z);
  }
  inline Real imag(Real x)
  {
    (void)(x);
    return Real(0);
  }

} // namespace hell

#endif // HELL_TYPES_HH
