#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctionsNew.hh"
#include "DIS.hh"
#include "DISinterpolator.hh"
#include "hHiggs_new_var.hh"
#include "QQbarPair.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "math/special_functions.hh"
#include "integration.hh"
#include "version.hh"
#include "h_coefficients.hh"
#include "expansion_AD_SF.hh"
//#include <complex_bessel/sp_airy.h>


using namespace std;



namespace HELL {

  unsigned int RCmodeNew = 0;
  void SetRCmodeCFnew(unsigned int mode) {  // mode=0 uses gammadot, mode=1 uses as*beta0
    if(mode>1) {
      cout << "Error: invalid RC mode. Using default." << endl;
      mode = 0;
    }
    RCmodeNew = mode;
  }
  bool RCvarNew = 0;
  void SetRCvarCFnew(bool var) {  // var=true uses "Aiy" anomalous dimension
    RCvarNew = var;
  }




  //
  // integrand of the form:  U'(x, xi, muF) * Coff(xi, pars)   or   - U(x, xi, muF) * dCoff/dxi(xi, pars)
  //
  double integrand(double xi, double x, xTableU *tU, process p, double asb0, double muFrat, void *par) {
    double dU = 0;
    double xiF = muFrat*muFrat;  // needed for the collinear subtractions
    double xiU = xi/xiF;  // use the variable ( Q^2/muF^2 * xi )
    double  U = tU->evalU(x, xiU, dU);
    /*
    if(xiU<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU/log(xiU);
      dU = dU * HTdamp + U * dHTdamp;
      U *= HTdamp;
    }
    dU /= xiF;    // convert derivatives of U wrt their arguments (xiU) to derivatives wrt xi
    //*/
    //
    //cout << x << "  " << xi << endl;
    double res = U;
    //
    if(p == gamma_qg) res *= - Dxi_Hqg(xi);
    //
    //if(p == CL) res *= - Dxi_CLoff(xi, x1);
    //if(p == C2) res *= - Dxi_C2off(xi, x1) - (xi<=xiF ? 1./xi : 0);
    if(p == CL) res *= - Dxi_CLoff(xi);
    if(p == C2) res *= - Dxi_C2off(xi) - (xi<=xiF ? 1./xi : 0);
    //if(p == CL) res = CLoff(xi) * dU;
    //if(p == C2) res = (C2off(xi) + (xi<=1 ? log(xi) : 0)) * Dxi_evolU(xi,gamma,r);
    //
    //if(p == Ctest1) res = - (xi<=1 ? 1. : 0) * U + (xi<=1 && xi>=0.9 ? 10. : 0);
    //if(p == Ctest2) res = (xi<=1 ? xi : 0) * Dxi_evolU(xi,gamma,r);
    //if(p == Ctest1) res = - Dxi_CLoff(xi) * U;
    //if(p == Ctest2) res = CLoff(xi) * Dxi_evolU(xi,gamma,r);
    //
    if(p>=50 && p<60) {
      double mQ = *(double*) par;
      if(p == Khg) res *= - Dxi_Khgoff(xi,mQ) + (xi<=xiF ? 1./xi : 0);
    }
    if(p >= 60 && p<100) {
      Interpolator *interp = (Interpolator*) par;
      if(p == MC2)   res *= - Dxi_MC2off(xi, interp);
      if(p == MCL)   res *= - Dxi_MCLoff(xi, interp);
      if(p == MC2ax) res *= - Dxi_MC2axoff(xi, interp);
      if(p == MCLax) res *= - Dxi_MCLaxoff(xi, interp);
      //
      double xim = interp->Getxim();
      if(p == MC2cc) res *= - Dxi_MC2CCoff(xi, interp) -               (xi<=xiF ? 1./xi : 0) /2;
      if(p == MCLcc) res *= - Dxi_MCLCCoff(xi, interp) - xim/(1+xim) * (xi<=xiF ? 1./xi : 0) /2;
      if(p == MC3cc) res *= - Dxi_MC3CCoff(xi, interp) -   1/(1+xim) * (xi<=xiF ? 1./xi : 0) /2;
    }
    if(p == ggHaux) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      res *= ggHsx->sf1y1dt(xi/2)/2;
    }
    if(p == DY) {
      res *= - Dxi_CDYoff(xi) - (xi<=xiF ? 1./xi : 0);
    }
    //
    //cout << setw(15) << xi << setw(15) << x1 << setw(35) << res << endl;
    return res;
  }


  //
  // integrand of the form:  U'(x1, xi1, muF) * U'(x2, xi2, muF) * Coff(xi1, xi2, pars)
  //
  //double integrand2d(double t, double y, double x1, double x2, xTableU *tU, process p, double asb0, double muFrat, void *par, double *intVars) {
  double integrand2d(double xi1, double xi2, double x1, double x2, xTableU *tU, process p, double asb0, double muFrat, void *par, double *intVars) {
    //double xi1 = t*(1+y);
    //double xi2 = t*(1-y);
    double xiF = muFrat*muFrat;
    double xiU1 = xi1/xiF;
    double xiU2 = xi2/xiF;
    /////////////////////////////////////////////////////
    double xi0 = exp(-1./asb0);
    if(xiU1<xi0) {
      if(fabs(xiU1-xi0)/xi0<1e-3) {
	xiU1 = xi0;
      } else {
	cout << "cazzo!  xi0=" << xi0 << "  xi1-xi0=" << xiU1-xi0 << endl;//<< "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU2<xi0) {
      if(fabs(xiU2-xi0)/xi0<1e-3) {
	xiU2 = xi0;
      } else {
	cout << "cazzo!  xi0=" << xi0 << "  xi2-xi0=" << xiU2-xi0 << endl;// << "   xi2="<<xiU2 << "   t="<<t << " y="<<y << endl;
	return 0;
      }
    }
    if(xiU1<xi0 || xiU2<xi0) return 0;
    /////////////////////////////////////////////////////
    //
    double res = 0;
    //
    if(p == ggH) {
      double dU1, U1 = tU->evalU(x1, xiU1, dU1);
      double dU2, U2 = tU->evalU(x2, xiU2, dU2);
      double t = (xi1+xi2)/2.;
      double y = (xi1-xi2)/(xi1+xi2);
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      //res = 4. * dU1 * dU2 * t * ( xi1*xi2*ggHsx->sf2(t,y) + ggHsx->sf1(t,y) );
      res = 4. * dU1 * dU2 * t * xi1*xi2*ggHsx->sf2(t,y) - U1*U2*2.*y*ggHsx->sf1dtdy(t,y) - (U1*U2+U2*xi1*dU1+U1*xi2*dU2)*ggHsx->sf1dt(t,y) + (U2*dU1-U1*dU2)*(1-y*y)*ggHsx->sf1dy(t,y);
      //res = 4. * dU1 * dU2 * t * xi1*xi2*ggHsx->sf2(t,y) - U1*U2*y*ggHsx->sf1dtdy(t,y) - (U1*U2+U2*xi1*dU1+U1*xi2*dU2)*ggHsx->sf1dt(t,y) + (U2*xi1*dU1-U1*xi2*dU2)/t*ggHsx->sf1dy(t,y);
      res /= 2*t; // this is to compensate for the Jacobian that is now included in the other routine
      if(res!=res) {
	cout << setw(18) << t << setw(12) << y << setw(12) << x1 << setw(12) << x2 << setw(12) << x1*x2;
	cout << setw(18) << U1 << setw(14) << xiU1*dU1 << setw(14) << U2 << setw(14) << xiU2*dU2;
	cout << setw(18) << ggHsx->sf1dtdy(t,y) << setw(14) << ggHsx->sf1dt(t,y) << setw(14) << ggHsx->sf1dy(t,y);
	cout << endl;
      }
    }
    //
    return res;
  }
  struct s_prdtv { process p; double d1, d2, d3, d4, d5; xTableU *tU; void *par; };
  double integrandGSLx(double x, void *pars) {
    s_prdtv p = *(s_prdtv*)pars;
    double xx  = p.d1;
    //xTableU *tU = p.tU;
    double a0 = p.d2;  // as(Q^2)
    double b0 = p.d3;  // beta0
    double muFrat = p.d4;
    //double lFQ = 2*log(muFrat);
    double xiF = muFrat*muFrat;
    //double xi0 = real(r); // may be negative - problematic
    //double xi0 = a0*b0;
    //if(real(r)>0 && real(r)>xi0) xi0 = real(r); // not sure if it makes sense...
    //double xi = exp(-1./a0/b0) + (1.-x)/x;  double jac = 1./x/x;
    //double mult = 3; double t = (mult*x-1)/a0/b0; double xi = exp(t);  double jac = xi*mult/a0/b0;
    //double mult = 3; double u0=1./a0/b0-lFQ; double t = x*u0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1*u0; double jac2 = mult*xi2*u0;
    //double mult = 3; if(a0<0.007) mult=2; double u01=1./a0/b0-lFQ; double t1 = x*u01; double u02=1./a0/b0; double t2 = x*u02; double xi1 = exp(-t1); double xi2 = exp(mult*t2); double jac1 = xi1*u01; double jac2 = mult*xi2*u02;
    double mult = 3; if(a0<0.007) mult=2; double u0=1./a0/b0; double t = x*u0; double xi1 = xiF*exp(-t); double xi2 = xiF*exp(mult*t); double jac1 = xi1*u0; double jac2 = mult*xi2*u0;
    //double mult = 3; double u=x*x; double t = u/a0/b0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1/a0/b0*2*x; double jac2 = mult*xi2/a0/b0*2*x;
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x;  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x;  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    //double Xmin = 0.05, Xmax = 70;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x;  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //cout << x << "  " << xi << endl;
    //cout << xi1 << "   " << xi2 << endl;
    //cout << u01 << "  " << lFQ << endl;
    //if(1+real(r)*log(xi1)<0) cout << "warning! xi1="<< xi1 << "  r="<< r << endl;
    //if(1+real(r)*log(xi2)<0) cout << "warning! xi2="<< xi2 << "  r="<< r << endl;
    //
    //cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi2 << "  " << x << endl;
    //if(x>1e-5) cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi << "  " << x << endl;
    //double result = jac  * integrand(xi,  xx, p.tU, p.p, a0*b0, muFrat, p.par);
    double result = jac1 * integrand(xi1, xx, p.tU, p.p, a0*b0, muFrat, p.par) + jac2 * integrand(xi2, xx, p.tU, p.p, a0*b0, muFrat, p.par);
    //if(result==result) cout << result << endl;
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    //delete tU;
    return result;
  }
  int integrandCUBAx(const int *ndim, const double x[], const int *ncomp, double res[], void *pars, int whichFunction) {
    //if(x[1]==0)
    s_prdtv p = *(s_prdtv*)pars;
    double xx  = p.d1;
    double xz  = xx;
    double a0 = p.d2;  // as(Q^2)
    double b0 = p.d3;  // beta0
    double muFrat = p.d4;
    double rapY  = p.d5;
    //double lFQ = 2*log(muFrat);
    double xiF = muFrat*muFrat;
    void *par = p.par;
    process pr = p.p;
    double xi1a, xi2a, xi1b, xi2b, jac1=1, jac2=1;
    //
    double *intVars = new double[*ndim];
    for(int i=0; i<*ndim; i++) intVars[i] = x[i];
    //
    double u0=1./a0/b0; double xi0=xiF*exp(-u0); double u=x[0]*u0; double mult=1.2;
    //double t1=xiF*exp(-u);     jac1*=t1*u0;      double y1=(1-xi0/t1)*(1.-x[1]*x[1]); jac1*=(1-xi0/t1)*2*x[1];
    double t1=xiF*exp(-u);     jac1*=t1*u0;      double y1=(1-xi0/t1)*x[1]; jac1*=(1-xi0/t1);
    //double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2=(1-xi0/t2)*(1.-x[1]*x[1]); jac2*=(1-xi0/t2)*2*x[1]; // is it better to unlink the two integration variables?
    double t2=xiF*exp(mult*u); jac2*=mult*t2*u0; double y2=(1-xi0/t2)*x[1]; jac2*=(1-xi0/t2); // is it better to unlink the two integration variables?
    xi1a = t1*(1+y1);  xi2a = t1*(1-y1);  jac1*=2*t1;
    xi1b = t2*(1+y2);  xi2b = t2*(1-y2);  jac2*=2*t2;
    //
    if(pr == ggH) { // inclusive
      rapY  = -log(xx) * (x[2]-1./2.);
      jac1 *= -log(xx);
      jac2 *= -log(xx);
    }
    //
    //cout << rapY << endl;
    // these are the x arguments of the evolutors U
    double xx1  = sqrt(xz)*exp( rapY);
    double xx2  = sqrt(xz)*exp(-rapY);
    //cout << xx1 << "  " << xx2 << endl;
    //
    // if(pr == ggH) { // inclusive ---- ALTERNATIVE IMPLEMENTATION
    //   xx1  = xx+(1-xx)*x[2];
    //   xx2  = xx/xx1;
    //   jac1 *= (1-xx)/xx1;
    //   jac2 *= (1-xx)/xx1;
    // }
    //
    if(whichFunction == 0) {
      res[0] = jac1 * integrand2d(xi1a, xi2a, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars) + jac2 * integrand2d(xi1b, xi2b, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
      //res[0] = jac1 * integrand2d(t1, y1, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars) + jac2 * integrand2d(t2, y2, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    } else if(whichFunction == 1) {
      res[0] = jac1 * integrand2d(xi1a, xi2a, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
      //res[0] = jac1 * integrand2d(t1, y1, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    } else if(whichFunction == 2) {
      res[0] = jac2 * integrand2d(xi1b, xi2b, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
      //res[0] = jac2 * integrand2d(t2, y2, xx1, xx2, p.tU, pr, a0*b0, muFrat, par, intVars);
    }
    //
    if(res[0]!=res[0]) {
      res[0]=0;  // tappullo!!
      /*
      cout << whichFunction << setw(12) << x[0] << setw(12) << x[1] << setw(12) << x[2];
      cout << setw(18) << t1 << setw(12) << y1 << setw(12) << t2 << setw(12) << y2 << setw(12) << rapY << setw(12) << xx1 << setw(12) << xx2 << setw(12) << xx;
      cout << endl;
      //*/
    } else {
      //cout << x[0] << "   " << x[1] << "   " << x[2] << "   " << res[0] << endl;
    }
    return 0;
  }
  int integrandCUBAx0(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 0);
  }
  int integrandCUBAx1(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 1);
  }
  int integrandCUBAx2(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    return integrandCUBAx(ndim, x, ncomp, res, pars, 2);
  }


  double integral1d_DeltaFRes(s_prdtv intpars) {
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    int status;
    double res = integration_gauss(integrandGSLx, 0, 1, prec, NULL, &intpars, &status);
    if(status) res = 0;
    gsl_set_error_handler (old_handler);
    return res;
  }
  double integralDd_DeltaFRes(s_prdtv intpars, int dim) {
    const int ncomp = 1;
    double cres[ncomp], error[ncomp], prob[ncomp];
    double epsrel=1e-2, epsabs=1.e-10;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    int mineval = 0, maxeval = 300000*pow(5,dim);
    //int mineval = 5000, maxeval = 1000000;
    //int nstart = 50000, nincrease = 20000, nbatch = 20000, gridno = 1;
    //int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 5;
    int nstart = 40000, nincrease = 40000, nbatch = 1000, gridno = 5;
    nregions=1; // to remove the warning
    int seed = 0;
    double result = 0;
    //int cores = 1, pmax= 10000;
    //cubacores(&cores,&pmax);
    //
    //
    //Vegas(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
    //Suave(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    //Cuhre(dim, ncomp, integrandCUBAx0, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    //result = cres[0];
    //
    //Cuhre(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    Vegas(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
    //Suave(dim, ncomp, integrandCUBAx1, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
    result += cres[0];
    //
    if(intpars.p != QQbarPairM2Yq && intpars.p != QQbarPairM2Yq_v2 && intpars.p != QQbarPairM2Y_v2 && intpars.p != QQbarSingleQuarkYq && intpars.p != QQbarSingleQuarkYq_v2) {
      cout << "ciccio" << endl;
      //Cuhre(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      Vegas(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
      //Suave(dim, ncomp, integrandCUBAx2, &intpars, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 1, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
      result += cres[0];
    }
    //
    if(fail == -1)  cout << "  ERROR: dimension out of range." << endl;
    else if(fail>0) cout << "  ERROR: precision not reached." << endl;
    //cout << result << " +/- " << error[0] << endl;
    return result;
  }

  double newDeltaFRes(double x, double as, int nf, xTableU *tU, process proc, int ordersub=0, double muFrat=1, void *pars=NULL, double rapY=0) {
    double b0 = beta0(nf);
    //
    s_prdtv intpars = { proc, x, as, b0, muFrat, rapY, tU, pars };
    double res = 0;
    if(proc == ggH)
      res = integralDd_DeltaFRes(intpars, 3);
    else
      res = integral1d_DeltaFRes(intpars);
    //
    //
    //double dU = 0;
    //double  U = tU->eval(x, exp(-1./as/b0), dU);
    //res += U*CLoff(exp(-1./as/b0));
    //
    hCoeff hC(proc, muFrat, pars);
    //double h0 = hC.hk(0);
    //
    double xP0 = x*P0(x,nf);
    //double P1 = P1sing(x,nf);
    //if(ordersub>0) res -= h0;  // this is a delta function in z=1 -> ignore it
    if(proc == ggH) {
      double h10 = hC.hk(1);
      double h20 = hC.hk(2);
      double h11 = hC.hk(3);
      double h30 = hC.hk(4);
      double h21 = hC.hk(5);
      if(ordersub>1) res -= 2*h10*as*xP0;
      if(ordersub>2) {
	double xP0sq = x*P0sq(x,nf);
	double xP1   = x*P1(x,nf);
	res -= as*as*(2*h20*(xP0sq-b0*xP0) + 2*h10*xP1 + h11*xP0sq);
	if(ordersub>3) {
	  double xP0cb = x*P0cb(x,nf);
	  double xP0P1 = x*P0P1(x,nf);
	  double xP2   = x*P2(x,nf,RCvarNew);
	  double T = (RCmodeNew==1 ? 1 : 2);
	  res -= as*as*as * ( 2*h30*(xP0cb-3*b0*xP0sq+2*b0*b0*xP0) + 2*h20*(2*xP0P1-T*b0*xP1) + 2*h10*xP2 + 2*h21*(xP0cb-b0*xP0sq) + 2*h11*xP0P1 );
	}
      }
    } else {
      double h1 = hC.hk(1);
      double h2 = hC.hk(2);
      double h3 = hC.hk(3);
      if(ordersub>1) res -= h1*as*xP0;
      if(ordersub>2) {
	double xP0sq = x*P0sq(x,nf);
	double xP1   = x*P1(x,nf);
	res -= as*as*(h2*(xP0sq-b0*xP0) + h1*xP1); // for DIS coefficient functions matched to NNLO we don't need this
	if(ordersub>3) {
	  double xP0cb = x*P0cb(x,nf);
	  double xP0P1 = x*P0P1(x,nf);
	  double xP2   = x*P2(x,nf,RCvarNew);
	  double T = (RCmodeNew==1 ? 1 : 2);
	  res -= as*as*as * ( h3*(xP0cb-3*b0*xP0sq+2*b0*b0*xP0) + h2*(2*xP0P1-T*b0*xP1) + h1*xP2 );
	}
      }
    }
    if(ordersub>4) cout << "Error, trying to subtract beyond the capability of HELL" << endl;
    //
    return res;
  }


  double DeltaUqgRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = tU->evalUqg(x);
    return res;
  }

  // matching condition
  double DeltaKhgRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, Khg, 2, muFrat, pars);
    res *= as/3./M_PI;
    res += tU->evalUqg(x)/nf;
    return res;
  }

  // massless DIS Coeff Functions
  double DeltaCLRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, CL, 2, muFrat);
    return res * as*nf/3./M_PI;
  }
  double DeltaC2Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, C2, 2, muFrat);
    res *= as*nf/3./M_PI;
    res -= tU->evalUqg(x);
    return res;
  }

  // massive DIS Coeff Functions in Neutral Current
  double DeltaMC2Res(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  double DeltaMCLRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MCL, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  // axial contributions
  double DeltaMC2ResAxial(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2ax, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  double DeltaMCLResAxial(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MCLax, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }

  // massive DIS Coeff Functions in Charged Current
  double DeltaMC2ResCC(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2cc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below? if not, it has to be applied before subtraction of the singularity, in integrand()
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2.;
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }
  double DeltaMCLResCC(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    double X = 1/(1+xim);
    if(x>=X) return 0;    
    double res = newDeltaFRes(x, as, nf, tU, MCLcc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below?
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2. * xim/(1+xim);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }
  double DeltaMC3ResCC(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    double X = 1/(1+xim);
    if(x>=X) return 0;    
    double res = newDeltaFRes(x, as, nf, tU, MC3cc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below?
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2. /(1+xim);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }

  
  double DeltaHggAuxRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, ggHaux, 4, muFrat, pars);
    return res;
  }

  double DeltaHggRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, ggH, 4, muFrat, pars);
    res -= 2*newDeltaFRes(x, as, nf, tU, ggHaux, 1, muFrat, pars); // adds back the delta contributions
    return res;
  }


  // Drell-Yan inclusive
  double DeltaCDYRes(double x, double as, int nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, DY, 2, muFrat, pars);
    res *= as/6./M_PI;
    res -= tU->evalUqg(x)/nf/2.;
    return res;
  }



};

