#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>

#include "Pqg.hh"
#include "math/borel.hh"
#include "commons.hh"
#include "gammaNLO.hh"
#include "h_coefficients.hh"
#include "CoefficientFunctions.hh"
#include "expansion_AD_SF.hh"


using namespace std;



namespace HELL {


  unsigned int RCmodeOld = 0;
  void SetRCmodeCFold(unsigned int mode) {  // mode=0 uses gammadot, mode=1 uses as*beta0
    if(mode>1) {
      cout << "Error: invalid RC mode. Using default." << endl;
      mode = 0;
    }
    RCmodeOld = mode;
  }
  bool RCvarOld = 0;
  void SetRCvarCFold(bool var) {  // var=true uses "Aiy" anomalous dimension
    RCvarOld = var;
  }

  dcomplex hRC(dcomplex gamma, dcomplex gammadot, hCoeff hC, int nPade=8) {
    int Ncoeff = hC.hk().size();
    if(Ncoeff < 2*nPade) {
      cout << "Not enough coefficients" << endl;
      exit(345);
    }
    Ncoeff = 2*nPade;
    //
    dcomplex *coeff_ck   = new dcomplex[Ncoeff];
    //dcomplex *sqbr_gamma = new dcomplex[Ncoeff];
    dcomplex gammadot_over_gamma2 = gammadot/gamma/gamma;
    //sqbr_gamma[0] = 1.;
    dcomplex sqbr_gamma = 1.;
    coeff_ck[0] = hC.hk(0);
    for(int k=1; k<Ncoeff; k++) {
      //sqbr_gamma[k] = gamma * ( 1. + (k-1.)*gammadot_over_gamma2 ) * sqbr_gamma[k-1];
      //coeff_ck[k] = hC.hk(k) * sqbr_gamma[k];
      sqbr_gamma *= gamma * ( 1. + (k-1.)*gammadot_over_gamma2 );
      coeff_ck[k] = hC.hk(k) * sqbr_gamma;
      //cout << coeff_ck[k] << endl;
    }
    //double Lambda = 20.;
    //nPade = Ncoeff/2-1;
    SetBorelOrder(2);  // !!!!!! <<=== parameter here: works with 1 or 2 
    //dcomplex borelsum = BorelPadeSum(1., nPade, coeff_ck, Ncoeff);
    dcomplex borelsum = BorelPadeSum(1., nPade, nPade-1, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelPadeSum(1., nPade, 1, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumCUBA (1., hqgNcoeff-1, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumGauss(1., hqgNcoeff-1, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumCUBA (1., 20, Lambda, coeff_ck, Ncoeff);
    //dcomplex borelsum = BorelTruncatedSumGauss(1., 13, Lambda, coeff_ck, Ncoeff); // Peraro like
    return borelsum;
  }



  dcomplex DeltaFRes(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, process proc, int ordersub=0, int nPade=8, void *pars=NULL) {
    hCoeff hC(proc, 1);
    double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    double h2 = hC.hk(2);
    double h3 = hC.hk(3);
    //
    double b0 = beta0(nf);
    dcomplex gamma0 = gamma0sing(N,nf);
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    //
    //gamma0 = gamma0(N,nf) - gamma0(N+1.,nf);
    //gamma0 = gamma0(N,nf) - 2.*gamma0(N+1.,nf) + gamma0(N+2.,nf);
    //gamma0 = gamma0(N,nf) - 3.*gamma0(N+1.,nf) + 3.*gamma0(N+2.,nf) - gamma0(N+3.,nf);
    //*
    // Marco version
    dcomplex gamma    = Delta_gamma    + as*gamma0 + as*as*gamma1 + as*as*as*gamma2;
    dcomplex gammadot = Delta_gammadot +    gamma0 + 2.*as*gamma1 + 3.*as*as*gamma2;
    //
    //gammaLO_BatNLO /= N+1.; // hack! large-N damping
    //gammaLOdot_BatNLO /= N+1.; // hack! large-N damping
    if(RCmodeOld == 1) gammadot = gamma/as;
    //
    gammadot *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    //gammaLOdot_BatNLO = 0;
    dcomplex res = 0;
    if(proc == U) {
      double xi = *(double*)pars;
      dcomplex r = -gammadot / gamma;
      //res = evolU_nodamp(xi, gamma, r)-1.;   // this is U-1, as the 1 gives a delta(1-z) contribution after Mellin inversion
      res = evolU(xi, gamma, r, as*b0)-1.;   // this is U-1, as the 1 gives a delta(1-z) contribution after Mellin inversion
      //res = evolU(xi, as*CA/N/M_PI, as*b0, as*b0)-1.;   // hack to reproduce the toy evolutor (whose inverse Mellin can be computed analytically)
      //res = exp(as*CA/N/M_PI*log(xi))-1.;   // hack to reproduce the toy evolutor (whose inverse Mellin can be computed analytically)
    } else if(proc == dU) {
      double xi = *(double*)pars;
      dcomplex r = -gammadot / gamma;
      res = Dxi_evolU(xi, gamma, r, as*b0);
      //res = Dxi_evolU_nodamp(xi, gamma, r);
    } else {
      res = hRC(gamma, gammadot, hC, nPade);
    }
    /*/
    // Richard version // not yet modified to account for the fact that Delta is now Delta2!!
    dcomplex FOgammaLO  = as*gamma0pl(N+1.,0);
    dcomplex gammaLO_BatNLO    = Delta_gammaLO_BatNLO    + FOgammaLO;
    dcomplex gammaLOdot_BatNLO = Delta_gammaLOdot_BatNLO + FOgammaLO/as;
    if(RCmodeOld == 1) gammaLOdot_BatNLO = gammaLO_BatNLO/as;
    gammaLOdot_BatNLO *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    FOgammaLO -= as*gamma0sing(N,0);
    dcomplex FOgammaLOdot  = -b0*as*FOgammaLO;
    //gammaLOdot_BatNLO = 0;
    //FOgammaLOdot = 0;
    dcomplex res = hRC(gammaLO_BatNLO, gammaLOdot_BatNLO, hC) - hRC(FOgammaLO, FOgammaLOdot, hC) + h0;
    //*/
    //
    //cout << res << endl;
    //if(res != 1.) exit(0);
    //
    //if(useLLp) gamma2 = gamma2sing(N,nf,RCvarOld); // recompute it since it was set to zero before
    double T = (RCmodeOld==1 ? 1 : 2);
    if(ordersub>0) res -= h0;
    if(ordersub>1) res -= h1*as*gamma0;
    if(ordersub>2) res -= as*as*(h2*gamma0*(gamma0-b0) + h1*gamma1);
    if(ordersub>3) res -= as*as*as*(h3*gamma0*(gamma0-b0)*(gamma0-2*b0) + h2*gamma1*(2.*gamma0-T*b0) + h1*gamma2);
    if(ordersub>4) cout << "Error, trying to subtract beyond the capability of HELL" << endl;
    //
    if(proc == logR) res *= gammadot/gamma;
    //
    //cout << setw(15) << real(gammaLO_BatNLO) << setw(16) << real(res) << endl << endl;
    return res;
  }

  //dcomplex Delta1Gamma(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
  dcomplex Delta1Gamma(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return Delta_gamma + as*as*gamma1 + as*as*as*gamma2;
  }
  //dcomplex Delta1GammaDot(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
  dcomplex Delta1GammaDot(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    //dcomplex gamma2 = (useLLp ? 0 : gamma2sing(N,nf,RCvarOld) );
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return Delta_gammadot + 2.*as*gamma1 + 3.*as*as*gamma2;
  }
  //dcomplex Gamma0(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
  dcomplex Gamma0(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0;
  }
  dcomplex Gamma0sq(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0*gamma0;
  }
  dcomplex Gamma0cb(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    return gamma0*gamma0*gamma0;
  }
  dcomplex Gamma1(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma1 = gamma1sing(N,nf);
    return gamma1;
  }
  dcomplex Gamma0Gamma1(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma0 = gamma0sing(N,nf);
    dcomplex gamma1 = gamma1sing(N,nf);
    return gamma0*gamma1;
  }
  dcomplex Gamma2(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex gamma2 = gamma2sing(N,nf,RCvarOld);
    return gamma2;
  }

  dcomplex DeltaGammaResNLOqg(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, gamma_qg, 3, 8);
    return res * as*nf/3./M_PI;
  }
  dcomplex DeltaUqg(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, Uqg, 1, 7); // should be 2, but then it's doubly counted in C2, so it's 1
    //dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, Uqg, 0, 7);
    return res * as*nf/3./M_PI;
  }
  dcomplex DeltaUxi(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, U, 1, 7, pars);
    double xi = *(double*)pars;
    return res;
  }
  dcomplex DeltaDUxi(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, dU, 1, 7, pars);
    double xi = *(double*)pars;
    return res;
  }
  dcomplex DeltaLogR(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, logR, 3, 7);
    return res; // any factor?
  }
  /*
  dcomplex DeltaGammaPlusUqg(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, dcomplex Delta_gammaNLO, void *pars) {
    double b0 = beta0(nf);
    double c = (11*CA+2*nf*(1-2*CF/CA))/12.;
    dcomplex fixedorderpole = ( CA/N - c*fmom(N+1.) )/M_PI;
    dcomplex gamma    = Delta_gamma    + as*fixedorderpole;
    dcomplex gammadot = Delta_gammadot +    fixedorderpole;
    //gammadot = gamma/as; //hack!!!
    gammadot *= -b0*as*as;  // convert deriv wrt as to deriv wrt t
    dcomplex r = -gammadot / gamma;
    dcomplex gamma0 = fixedorderpole;
    dcomplex gamma1 = nf*(26.*CF-23.*CA)/36./M_PI/M_PI/N;
    dcomplex gammaNLO = as*gamma0 + as*as*gamma1 + Delta_gammaNLO;
    hCoeff hC(Uqg,1);
    double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    //dcomplex extraU  =  h0*as/(r + gamma);
    dcomplex extraU  =  h0*as/(r + gammaNLO);
    dcomplex extraU0 =  h0/(b0+gamma0);
    dcomplex extraU1 = -h0*gamma1/gamma0/(b0+gamma0);
    extraU = extraU0 = extraU1 = 0; // hack!
    dcomplex U  = nf/3./M_PI * ( as*DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, Uqg, 0, 7) + extraU );
    dcomplex U0 = nf/3./M_PI * ( extraU0 );
    dcomplex U1 = nf/3./M_PI * ( extraU1 + h1 );
    return gammaNLO*U - as*gamma0*U0 - as*as*(gamma1*U0 + gamma0*U1);
    return (gammaNLO+r)*U - as*(gamma0+b0)*U0 - as*as*(gamma1*(1.+b0/gamma0)*U0 + (gamma0+b0)*U1);
  }
  */
  dcomplex oldDeltaCLRes(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, CL, 2, 8);
    return res * as*nf/3./M_PI;
  }
  dcomplex oldDeltaC2Res(dcomplex Delta_gamma, dcomplex Delta_gammadot, dcomplex N, double as, double nf, void *pars) {
    dcomplex res = DeltaFRes(Delta_gamma, Delta_gammadot, N, as, nf, C2, 2, 8);
    return res * as*nf/3./M_PI;
  }



};

